/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "gl_local.h"
#include "keys.h"
#include "input.h"

#include <GLES/gl.h>
#include <GLES/egl.h>

#if defined(PANDORA)
#include "pnd_event.h"
#include "X11/Xlib.h"
#include "X11/Xutil.h"
#include "X11/Xatom.h"
#else
    #include <unistd.h>
    #include <fcntl.h>

    void WIZCAANOO_OpenInput( void );
    void WIZCAANOO_CloseInput( void );
    void WIZCAANOO_PollInput( void );

    #if defined(CAANOO)
    #include <linux/input.h>
    void ReadEvents( int fd );
    #else
    #define GPH_UP                          (1<<0)
    #define GPH_UP_LEFT                     (1<<1)
    #define GPH_LEFT                        (1<<2)
    #define GPH_DOWN_LEFT                   (1<<3)
    #define GPH_DOWN                        (1<<4)
    #define GPH_DOWN_RIGHT                  (1<<5)
    #define GPH_RIGHT                       (1<<6)
    #define GPH_UP_RIGHT                    (1<<7)
    #endif
    #define GPH_START                       (1<<8)          // Caanoo HOME
    #define GPH_SELECT                      (1<<9)          // Cannoo HELP2
    #define GPH_L                           (1<<10)
    #define GPH_R                           (1<<11)
    #define GPH_A                           (1<<12)
    #define GPH_B                           (1<<13)
    #define GPH_Y                           (1<<14)
    #define	GPH_X                           (1<<15)
    #define GPH_VOL_UP                      (1<<16)
    #define GPH_VOL_DOWN                    (1<<17)
    #define GPH_PUSH                        (1<<18)         // Caanoo only
    #define GPH_HELP1                       (1<<20)         // Caanoo only

    NativeWindowType hNativeWnd = 0;
    int stickfd=0;
    int gpiofd=0;
    int oldgpkey;
#endif

cvar_t	_windowed_mouse = {"_windowed_mouse", "1", CVAR_ARCHIVE};

static qbool mouse_avail = true;
static qbool mouse_active = false;
static int   mouse_x, mouse_y, mouse_buttonstate;
static int   old_mouse_x, old_mouse_y;
static int   mouse_oldbuttonstate = 0;

static cvar_t m_filter = {"m_filter", "0"};

static int scr_width, scr_height;

qbool gl_mtexable = false;
qbool gl_mtexfbskins = false;

#if defined(PANDORA)
int pandora_driver_mode = 0;

Window				x11Window	= 0;
Display*			x11Display	= 0;
long				x11Screen	= 0;
XVisualInfo			x11Visual;
XVisualInfo*		px11Visual	= 0;
Colormap			x11Colormap	= 0;
#endif

EGLDisplay      glDisplay;
EGLConfig       glConfig;
EGLContext      glContext;
EGLSurface      glSurface;

/*-----------------------------------------------------------------------*/


//static int SDL_LateKey(SDL_keysym *sym) {}


static void install_grabs(void) {}

static void uninstall_grabs(void)
{
	mouse_active = false;
}

void HotKey_ToggleFullScreen(void) {}

void HotKey_ToggleGrab(void) {}

void HotKey_Iconify(void) {}

void IN_DeactivateMouse( void )
{
#ifdef NOT_TESTING_WINDOWED
	if (!mouse_avail || !screen)
		return;
#endif

	if (mouse_active) {
		uninstall_grabs();
		mouse_active = false;
	}
}

void IN_ActivateMouse( void )
{
#ifdef NOT_TESTING_WINDOWED
	if (!mouse_avail || !screen)
		return;
#endif
	/* Com_Printf("IN_ActivateMouse entered\n"); */

	if (!mouse_active) {
		mouse_x = mouse_y = 0; // don't spazz
		install_grabs();
		mouse_active = true;
	}
}

void GL_CheckExtensions (void)
{
}

/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
	*x = *y = 0;
	*width = scr_width;
	*height = scr_height;

//	glViewport (*x, *y, *width, *height);
}


void GL_EndRendering (void)
{
	glFlush();
    eglSwapBuffers(glDisplay,glSurface);
}

EGLint attrib_list_fsaa[] =
{
  EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
  EGL_BUFFER_SIZE,  0,
  EGL_DEPTH_SIZE,   16,
  EGL_SAMPLE_BUFFERS, 1,
  EGL_SAMPLES,        4,
  EGL_NONE
};

EGLint attrib_list[] =
{
  EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
  EGL_BUFFER_SIZE,  0,
  EGL_DEPTH_SIZE,   16,
  EGL_NONE
};

void VID_Init(unsigned char *palette)
{
	int i;
#if defined(PANDORA)
	int width = 800, height = 480;
#else
	int width = 320, height = 240;
#endif

	Cvar_Register (&m_filter);
	Cvar_Register (&_windowed_mouse);

#if defined(PANDORA)
    PND_Setup_Controls();
#else
    WIZCAANOO_OpenInput();
#endif

	if ((i = COM_CheckParm("-width")) != 0)
		width = atoi(com_argv[i+1]);

	if ((i = COM_CheckParm("-height")) != 0)
		height = atoi(com_argv[i+1]);

	if ((i = COM_CheckParm("-conwidth")) != 0)
		vid.width = Q_atoi(com_argv[i+1]);
	else
		vid.width = 640;

	vid.width &= 0xfff8; // make it a multiple of eight

	if (vid.width < 320)
		vid.width = 320;

	// pick a conheight that matches with correct aspect
	vid.height = vid.width*3 / 4;

	if ((i = COM_CheckParm("-conheight")) != 0)
		vid.height = Q_atoi(com_argv[i+1]);
	if (vid.height < 200)
		vid.height = 200;

#if defined(PANDORA)
	enum
	{
	_NET_WM_STATE_REMOVE =0,
	_NET_WM_STATE_ADD = 1,
	_NET_WM_STATE_TOGGLE =2
	};

	Window                  sRootWindow;
	XSetWindowAttributes	sWA;
	unsigned int		    ui32Mask;
	int			            i32Depth;

	pandora_driver_mode = COM_CheckParm("-x11");

	if (pandora_driver_mode)
	{
		// Initializes the display and screen
		x11Display = XOpenDisplay( ":0" );
		if (!x11Display)
		{
			printf("Error: Unable to open X display\n");
		}
		x11Screen = XDefaultScreen( x11Display );

		// Gets the display parameters so we can pass the same parameters to the window to be created.
		sRootWindow	= RootWindow(x11Display, x11Screen);
		i32Depth	= DefaultDepth(x11Display, x11Screen);
		px11Visual	= &x11Visual;
		XMatchVisualInfo( x11Display, x11Screen, i32Depth, TrueColor, px11Visual);
		if (!px11Visual)
		{
			printf("Error: Unable to acquire visual\n");
		}
		// Colormap of the specified visual type for the display.
		x11Colormap = XCreateColormap( x11Display, sRootWindow, px11Visual->visual, AllocNone );
		sWA.colormap = x11Colormap;

		// List of events to be handled by the application. Add to these for handling other events.
		sWA.event_mask = StructureNotifyMask | ExposureMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask;

		// Display capabilities list.
		ui32Mask = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap;

		// Creates the X11 window
		x11Window = XCreateWindow( x11Display, RootWindow(x11Display, x11Screen), 0, 0, 800, 480,
									0, CopyFromParent, InputOutput, CopyFromParent, ui32Mask, &sWA);

		// Make the window viewable and flush the output buffer.
		XMapWindow(x11Display, x11Window);
		XFlush(x11Display);

		// Make the window fullscreen
		unsigned char fullScreen = 1;
		Atom wmState = XInternAtom(x11Display, "_NET_WM_STATE", False);
		Atom wmFullScreen = XInternAtom(x11Display,"_NET_WM_STATE_FULLSCREEN", False);

		XEvent xev;
		xev.xclient.type		= ClientMessage;
		xev.xclient.serial		= 0;
		xev.xclient.send_event		= True;
		xev.xclient.window		= x11Window;
		xev.xclient.message_type	= wmState;
		xev.xclient.format		= 32;
		xev.xclient.data.l[0]		= (fullScreen ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE);
		xev.xclient.data.l[1]		= wmFullScreen;
		xev.xclient.data.l[2]		= 0;

		XSendEvent(x11Display, DefaultRootWindow(x11Display), False, SubstructureRedirectMask | SubstructureNotifyMask, &xev);
	}
#else
    hNativeWnd = (NativeWindowType)malloc(16*1024);
    if(!hNativeWnd) {
        printf( "VID_Init: OS_CreateWindow Failed\n" );
        abort();
    }
#endif

	EGLint numConfigs;
	EGLint majorVersion;
	EGLint minorVersion;

#if defined(PANDORA)
	if (pandora_driver_mode) {
		glDisplay = eglGetDisplay( (NativeDisplayType)x11Display );
	} else
#endif
	{
		glDisplay = eglGetDisplay( (NativeDisplayType)0 );
	}

	if( glDisplay == EGL_NO_DISPLAY )
	{
		printf( "GL No Display failed\n" );
	}

	if( !eglInitialize( glDisplay, &majorVersion, &minorVersion ) )
	{
		printf( "GL Init failed\n" );
	}

	if( !eglChooseConfig( glDisplay, attrib_list, &glConfig, 1, &numConfigs ) )
	{
		printf( "GL Config failed\n" );
	}

	glContext = eglCreateContext( glDisplay, glConfig, NULL, NULL );
	if( glContext==0 )
	{
		printf( "GL Context failed\n" );
	}

#if defined(PANDORA)
	if (pandora_driver_mode) {
	      glSurface = eglCreateWindowSurface( glDisplay, glConfig, (NativeDisplayType)x11Window, NULL );
	} else {
	      glSurface = eglCreateWindowSurface( glDisplay, glConfig, (NativeDisplayType)0, NULL );
	}
#else
    glSurface = eglCreateWindowSurface( glDisplay, glConfig, hNativeWnd, NULL );
#endif

	if( glSurface==0 )
	{
		printf( "GL Surface failed\n" );
	}
	printf( "EGL Init Completed\n" );

	eglMakeCurrent( glDisplay, glSurface, glSurface, glContext );

	scr_width = width;
	scr_height = height;
	vid.realwidth = width;
	vid.realheight = height;

	if (vid.height > height)
		vid.height = height;
	if (vid.width > width)
		vid.width = width;
	vid.width = vid.width;
	vid.height = vid.height;

	vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);
	vid.numpages = 2;

	GL_Init();

	Com_Printf ("Video mode %dx%d initialized.\n", width, height);

	SCR_InvalidateScreen ();
}

void VID_Shutdown(void)
{
	IN_DeactivateMouse();

#if defined(PANDORA)
	PND_Close_Controls();
#else
    WIZCAANOO_CloseInput();
#endif

	eglMakeCurrent( glDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
	eglDestroySurface( glDisplay, glSurface );
	eglDestroyContext( glDisplay, glContext );
	eglTerminate( glDisplay );

#if defined(PANDORA)
	if (pandora_driver_mode) {
		if (x11Window) XDestroyWindow(x11Display, x11Window);
		if (x11Colormap) XFreeColormap( x11Display, x11Colormap );
		if (x11Display) XCloseDisplay(x11Display);
	}
#endif
}


void Sys_SendKeyEvents(void)
{
}

#if 0
void Force_CenterView_f (void)
{
	cl.viewangles[PITCH] = 0;
}
#endif

void IN_Init(void)
{
	if ( COM_CheckParm ("-nomouse") )
		return;
	mouse_x = mouse_y = 0;
	mouse_avail = 1;
}

void IN_Shutdown(void)
{
	mouse_avail = 0;
}

/*
===========
IN_Commands
===========
*/
void IN_Commands (void)
{
	int i;
	int mouse_buttonstate;

#if defined(PANDORA)
    PND_SendAllEvents( &mouse_x, &mouse_y, &mouse_buttonstate );
#else
    WIZCAANOO_PollInput();
#endif

	if (!mouse_avail) return;

	/* Quake swaps the second and third buttons */
	for (i=0 ; i<3 ; i++) {
		if ( (mouse_buttonstate & (1<<i)) && !(mouse_oldbuttonstate & (1<<i)) )
			Key_Event (K_MOUSE1 + i, true);

		if ( !(mouse_buttonstate & (1<<i)) && (mouse_oldbuttonstate & (1<<i)) )
			Key_Event (K_MOUSE1 + i, false);
	}

	if ( (mouse_buttonstate & (1<<3)) && !(mouse_oldbuttonstate & (1<<3)) )
		Key_Event (K_MWHEELUP, true);

	if ( (mouse_buttonstate & (1<<3)) && (mouse_oldbuttonstate & (1<<3)) )
		Key_Event (K_MWHEELUP, false);

	if ( (mouse_buttonstate & (1<<4)) && !(mouse_oldbuttonstate & (1<<4)) )
		Key_Event (K_MWHEELDOWN, true);

	if ( (mouse_buttonstate & (1<<4)) && (mouse_oldbuttonstate & (1<<4)) )
		Key_Event (K_MWHEELDOWN, true);

	mouse_oldbuttonstate = mouse_buttonstate;
}

/*
===========
IN_Move
===========
*/
void IN_Move (usercmd_t *cmd)
{
	if (!mouse_avail)
		return;

	if (m_filter.value)
	{
		mouse_x = (mouse_x + old_mouse_x) * 0.5;
		mouse_y = (mouse_y + old_mouse_y) * 0.5;
	}
	old_mouse_x = mouse_x;
	old_mouse_y = mouse_y;

	mouse_x *= sensitivity.value;
	mouse_y *= sensitivity.value;

    // add mouse X/Y movement to cmd
    if ( (in_strafe.state & 1) || (lookstrafe.value && mlook_active))
		cmd->sidemove += m_side.value * mouse_x;
	else
		cl.viewangles[YAW] -= m_yaw.value * mouse_x;

	if (mlook_active)
		V_StopPitchDrift ();

	if ( mlook_active && !(in_strafe.state & 1))
	{
		cl.viewangles[PITCH] += m_pitch.value * mouse_y;
		if (cl.viewangles[PITCH] > cl.maxpitch)
			cl.viewangles[PITCH] = cl.maxpitch;
		if (cl.viewangles[PITCH] < cl.minpitch)
			cl.viewangles[PITCH] = cl.minpitch;
	}
	else
	{
        cmd->forwardmove -= m_forward.value * mouse_y;
	}
	mouse_x = mouse_y = 0.0;
}

// direct draw software compatability stuff
void VID_LockBuffer(void) {}
void VID_UnlockBuffer(void) {}
void D_BeginDirectRect (int x, int y, byte *pbitmap, int width, int height) {}
void D_EndDirectRect (int x, int y, int width, int height) {}

void VID_SetCaption (char *text) {}

void VID_SetDeviceGammaRamp (unsigned short *ramps) {}


#if defined(CAANOO) || defined(WIZ)

void WIZCAANOO_OpenInput( void )
{
// GP2X/WIZ Button Control
#if defined(CAANOO)
     stickfd = open("/dev/input/event1", O_RDONLY|O_NONBLOCK);
     if (stickfd != 0)
         printf("Caanoo stick open\n");
     else
         printf("ERROR Caanoo stick NOT open\n");
#endif
#if defined(CAANOO) || defined(WIZ)
     gpiofd = open("/dev/GPIO", O_RDWR | O_NDELAY);
     if (gpiofd != 0)
         printf("GPH GPIO open\n");
     else
         printf("GPH GPIO  NOT open\n");
#endif
}

void WIZCAANOO_CloseInput( void )
{
#if defined(CAANOO) || defined(WIZ)
    close(gpiofd);
#endif
#if defined(CAANOO)
    close(stickfd);
#endif
}

#if defined(CAANOO)
#define DEAD_ZONE   64
#define ZERO_POINT 128

void ReadEvents( int fd )
{
    struct input_event event[16];
    int rbytes, i;

    rbytes = read(fd, &event, sizeof(event));
    if (rbytes <= 0)
        return;
    for (i=0; i<rbytes/sizeof(struct input_event); i++)
    {
        switch (event[i].type)
        {
            case EV_ABS:
                switch (event[i].code)
                {
                    case ABS_X:
                        if (event[i].value>ZERO_POINT+DEAD_ZONE)
                            Key_Event(K_RIGHTARROW, true);
                        else
                            Key_Event(K_RIGHTARROW, false);

                        if (event[i].value<ZERO_POINT-DEAD_ZONE)
                            Key_Event(K_LEFTARROW, true);
                        else
                            Key_Event(K_LEFTARROW, false);

                        printf("x = 0x%x\n", event[i].value);
                        break;
                    case ABS_Y:
                        if (event[i].value<ZERO_POINT-DEAD_ZONE)
                            Key_Event(K_UPARROW, true);
                        else
                            Key_Event(K_UPARROW, false);

                        if (event[i].value>ZERO_POINT+DEAD_ZONE)
                            Key_Event(K_DOWNARROW, true);
                        else
                            Key_Event(K_DOWNARROW, false);

                        printf("y = 0x%x\n", event[i].value);
                        break;
                    default   :
                        printf("type = 0x%x, code = 0x%x\n", event[i].type, event[i].code);
                }
                break;
            default:
                if (event[i].type)
                    printf("type = 0x%x, code = 0x%x\n", event[i].type, event[i].code);
                break;
        }
    }
}
#endif

void WIZCAANOO_PollInput( void )
{
  int gpkey;

  read(gpiofd, &gpkey, 4);

#if defined(CAANOO)
  ReadEvents( stickfd );
#else
  if ((gpkey&GPH_UP_LEFT) && !((gpkey&GPH_UP) || (gpkey&GPH_LEFT))) {
    gpkey=gpkey|GPH_UP;
    gpkey=gpkey|GPH_LEFT;
  }
  if ((gpkey&GPH_UP_RIGHT) && !((gpkey&GPH_UP) || (gpkey&GPH_RIGHT))) {
    gpkey=gpkey|GPH_UP;
    gpkey=gpkey|GPH_RIGHT;
  }
  if ((gpkey&GPH_DOWN_LEFT) && !((gpkey&GPH_DOWN) || (gpkey&GPH_LEFT))) {
    gpkey=gpkey|GPH_DOWN;
    gpkey=gpkey|GPH_LEFT;
  }
  if ((gpkey&GPH_DOWN_RIGHT) && !((gpkey&GPH_DOWN) || (gpkey&GPH_RIGHT))) {
    gpkey=gpkey|GPH_DOWN;
    gpkey=gpkey|GPH_RIGHT;
  }

  // UP
  if(gpkey & GPH_UP)        Key_Event(K_UPARROW, true);
  else if (oldgpkey & GPH_UP) Key_Event(K_UPARROW, false);
  // DOWN
  if(gpkey & GPH_DOWN)      Key_Event(K_DOWNARROW, true);
  else if (oldgpkey & GPH_DOWN) Key_Event(K_DOWNARROW, false);
  // LEFT
  if(gpkey & GPH_LEFT)      Key_Event(K_LEFTARROW, true);
  else if (oldgpkey & GPH_LEFT) Key_Event(K_LEFTARROW, false);
  // RIGHT
 if(gpkey & GPH_RIGHT)      Key_Event(K_RIGHTARROW, true);
  else if (oldgpkey & GPH_RIGHT)    Key_Event(K_RIGHTARROW, false);
#endif

  // ESC
  if(gpkey & GPH_START)	Key_Event(27, true);
  else if (oldgpkey & GPH_START) Key_Event(27, false);

  // CHANGE WEAPON (/)
  if(gpkey & GPH_X)	Key_Event(47, true);
  else if (oldgpkey & GPH_X) Key_Event(47, false);
  // RUN
  if(gpkey & GPH_Y)	Key_Event(K_SHIFT, true);
  else if (oldgpkey & GPH_Y) Key_Event(K_SHIFT, false);

  // CTRL / FIRE
  if(gpkey & GPH_A)	Key_Event(K_CTRL, true);
  else if (oldgpkey & GPH_A) Key_Event(K_CTRL, false);
  // Y / Yes
  if(gpkey & GPH_A)	Key_Event('y', true);
  else if (oldgpkey & GPH_A) Key_Event('y', false);
  // RETURN / JUMP
  if(gpkey & GPH_B)	Key_Event(K_ENTER, true);
  else if (oldgpkey & GPH_B) Key_Event(K_ENTER, false);

  // something
  if(gpkey & GPH_SELECT)	Key_Event('a', true);
  else if (oldgpkey & GPH_SELECT) Key_Event('a', false);

#if defined(CAANOO)
  if(gpkey & GPH_PUSH)        Key_Event('b', true);
  else if (oldgpkey & GPH_PUSH) Key_Event('b', false);

  if(gpkey & GPH_HELP1)        Key_Event('c', true);
  else if (oldgpkey & GPH_HELP1) Key_Event('c', false);
#endif

  if(gpkey & GPH_VOL_UP && gpkey & GPH_VOL_DOWN){
    Key_Event('a', true); // something
    if(oldgpkey & GPH_VOL_UP)
      Key_Event('d', false);
    if(oldgpkey & GPH_VOL_DOWN)
      Key_Event('c', false);
  }
  else if(oldgpkey & GPH_VOL_DOWN && oldgpkey & GPH_VOL_UP) Key_Event('a', false);
  else{
    if(gpkey & GPH_VOL_DOWN)	Key_Event('c', true); // swim down
    else if (oldgpkey & GPH_VOL_DOWN) Key_Event('c', false);
    if(gpkey & GPH_VOL_UP)	Key_Event('d', true); // swim up
    else if (oldgpkey & GPH_VOL_UP) Key_Event('d', false);
  }

  if(gpkey & GPH_L && gpkey & GPH_R){
	  Key_Event(K_END, true);
	  if(oldgpkey & GPH_L)
		  Key_Event(K_ALT, false);
	  if(oldgpkey & GPH_R)
		  Key_Event(K_INS, false);
  }
  else if(oldgpkey & GPH_L && oldgpkey & GPH_R) Key_Event(K_END, false);
  else{
	  if(gpkey & GPH_L)	Key_Event(K_ALT, true);
	  else if (oldgpkey & GPH_L) Key_Event(K_ALT, false);
	  if(gpkey & GPH_R)	Key_Event(K_INS, true);
	  else if (oldgpkey & GPH_R) Key_Event(K_INS, false);
  }

  oldgpkey=gpkey;
}

#endif /* defined(CAANOO) || defined(WIZ) */
